__version__ = '0.1.4.1'

STRTABLE = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
HKEY = b"3AF0298C219469522A313570E8583005A642E73EDD58E3EA2FB7339D3DF1597E"

_KEYPARAMS = {
	"osName": "android",
	"osVersion": "6.0.1",
	"osLand": "ja_JP",
	"osTimezone": "Asia/Tokyo",
	"appId": "tv.abema",
	"appVersion": "3.27.1"
}

_MEDIATOKEN_API = "https://api.abema.io/v1/media/token"
_LICENSE_API = "https://license.abema.io/abematv-hls"
_USERAPI = "https://api.abema.io/v1/users"
_PROGRAMAPI = 'https://api.abema.io/v1/video/programs/'
_CHANNELAPI = 'https://api.abema.io/v1/media/slots/'

def is_channel(url):
	url = url[:url.rfind('/')]
	url = url[url.rfind('/')+1:]
	if url == "slots":
		return True
	return False


import hmac
import hashlib
import struct
import time
import uuid
import requests
import m3u8 as M3U8
import re

from base64 import urlsafe_b64encode
from binascii import unhexlify
from Crypto.Cipher import AES

from .common import STRTABLE, HKEY, _MEDIATOKEN_API, _LICENSE_API, _USERAPI, _KEYPARAMS, _PROGRAMAPI, _CHANNELAPI, is_channel

def getAuthToken(session, verbose):
	def keySecret(devid):
		SECRETKEY = (b"v+Gjs=25Aw5erR!J8ZuvRrCx*rGswhB&qdHd_SYerEWdU&a?3DzN9B"
					b"Rbp5KwY4hEmcj5#fykMjJ=AuWz5GSMY-d@H7DMEh3M@9n2G552Us$$"
					b"k9cD=3TxwWe86!x#Zyhe")
		deviceid = devid.encode("utf-8")
		ts_1hour = (int(time.time()) + 60 * 60) // 3600 * 3600
		time_struct = time.gmtime(ts_1hour)
		ts_1hour_str = str(ts_1hour).encode("utf-8")
		
		h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
		h.update(SECRETKEY)
		tmp = h.digest()
		
		if verbose:
			print('[DEBUG] First salting data: {}'.format(tmp))

		for i in range(time_struct.tm_mon):
			h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
			h.update(tmp)
			tmp = h.digest()

		if verbose:
			print('[DEBUG] Second salting data: {}'.format(tmp))

		h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
		h.update(urlsafe_b64encode(tmp).rstrip(b"=") + deviceid)
		tmp = h.digest()

		if verbose:
			print('[DEBUG] Third salting data: {}'.format(tmp))

		for i in range(time_struct.tm_mday % 5):
			h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
			h.update(tmp)
			tmp = h.digest()

		if verbose:
			print('[DEBUG] Fourth salting data: {}'.format(tmp))

		h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
		h.update(urlsafe_b64encode(tmp).rstrip(b"=") + ts_1hour_str)
		tmp = h.digest()

		if verbose:
			print('[DEBUG] Fifth salting data: {}'.format(tmp))

		for i in range(time_struct.tm_hour % 5):  # utc hour
			h = hmac.new(SECRETKEY, digestmod=hashlib.sha256)
			h.update(tmp)
			tmp = h.digest()
		
		if verbose:
			print('[DEBUG] Last salting data: {}'.format(tmp))

		finalize = urlsafe_b64encode(tmp).rstrip(b"=").decode("utf-8")

		if verbose:
			print('[DEBUG] Key secret result: {}'.format(finalize))
			
		return finalize

	deviceid = str(uuid.uuid4())
	if verbose:
		print('[DEBUG] Generated Device UUID: {}'.format(deviceid))
	jsonData = {"deviceId": deviceid, "applicationKeySecret": keySecret(deviceid)}
	
	if verbose:
		print('[DEBUG] Sending json data')
	res = session.post(_USERAPI, json=jsonData).json()

	try:
		if verbose:
			print('[DEBUG] Data sended, getting token')
		token = res['token']
		if verbose:
			print('[DEBUG] Usertoken: {}'.format(token))
	except:
		print('[ERROR] Failed to get usertoken')
		import sys; sys.exit(1)
	
	return ['bearer ' + token, deviceid]

def fetchVideoKey(ticket=None, authToken=None, session=None, verbose=False):
	if verbose:
		print('[DEBUG] Sending parameter to API')
	restoken = session.get(_MEDIATOKEN_API, params=_KEYPARAMS).json()
	mediatoken = restoken['token']
	if verbose:
		print('[DEBUG] Mediatoken: {}'.format(mediatoken))

	if verbose:
		print('[DEBUG] Sending ticket and mediatoken to License API')
	gl = session.post(_LICENSE_API, params={"t": mediatoken}, json={"kv": "a", "lt": ticket}).json()

	cid = gl['cid']
	k = gl['k']

	if verbose:
		print('[DEBUG] CID: {}'.format(cid))
		print('[DEBUG] K: {}'.format(k))

	if verbose:
		print('[DEBUG] Summing up data with STRTABLE')
	res = sum([STRTABLE.find(k[i]) * (58 ** (len(k) - 1 - i)) for i in range(len(k))])

	if verbose:
		print('[DEBUG] Result: {}'.format(res))
		print('[DEBUG] Intepreting data')

	encvk = struct.pack('>QQ', res >> 64, res & 0xffffffffffffffff)

	if verbose:
		print('[DEBUG] Encoded video key: {}'.format(encvk))
		print('[DEBUG] Hashing data')

	h = hmac.new(unhexlify(HKEY), (cid + authToken[1]).encode("utf-8"), digestmod=hashlib.sha256)
	enckey = h.digest()

	if verbose:
		print('[DEBUG] Second Encoded video key: {}'.format(enckey))
		print('[DEBUG] Decrypting result')

	aes = AES.new(enckey, AES.MODE_ECB)
	vkey = aes.decrypt(encvk)

	if verbose:
		print('[DEBUG] Decrypted, Resulting output: {}'.format(vkey))
	
	return vkey

def parsem3u8(m3u8, session, verbose):
	if verbose:
		print('[DEBUG] Requesting m3u8')
	r = session.get(m3u8)
	if verbose and r.status_code == 200:
		print('[DEBUG] m3u8 requested')
		print('[DEBUG] Parsing m3u8')
	x = M3U8.loads(r.text)
	files = x.files
	iv = x.keys[0].iv
	ticket = x.keys[0].uri[18:]
	if verbose:
		print('[DEBUG] Total files: {}'.format(len(files)))
		print('[DEBUG] IV: {}'.format(iv))
		print('[DEBUG] Ticket key: {}'.format(ticket))
	return [files[1:], iv[2:], ticket]

def webparse_m3u8(m3u8, session, verbose):
	if verbose:
		print('[DEBUG] Requesting data to API')
	reg = re.compile('(program|slot)\/[\w+-]+')
	res = re.search(reg, m3u8)[0]
	eplink = res[res.find('/')+1:]

	if 'slot' in res:
		req = session.get(_CHANNELAPI + eplink)
		if verbose and req.status_code == 200:
			print('[DEBUG] Data requested')
			print('[DEBUG] Parsing json API')
		jsdata = req.json()
		title = jsdata['slot']['title']
		res = m3u8[:m3u8.rfind('/')]
		res = res[res.rfind('/')+1:]

		if verbose:
			print('[DEBUG] M3U8 Link: {}'.format(m3u8))
			print('[DEBUG] Title: {}'.format(title))
		
		return title, res
	else:
		req = session.get(_PROGRAMAPI + eplink)
		if verbose and req.status_code == 200:
			print('[DEBUG] Data requested')
			print('[DEBUG] Parsing json API')
		jsdata = req.json()
		title = jsdata['series']['title']
		epnum = jsdata['episode']['title']
		outputfile = title + ' - ' + epnum

		res = m3u8[:m3u8.rfind('/')]
		res = res[res.rfind('/')+1:]
		
		if verbose:
			print('[DEBUG] M3U8 Link: {}'.format(m3u8))
			print('[DEBUG] Video title: {}'.format(title))
			print('[DEBUG] Episode number: {}'.format(epnum))
			
		return outputfile, res

def webparse(url, res, session, verbose):
	if verbose:
		print('[DEBUG] Requesting data to API')
	eplink = url[url.rfind('/')+1:]

	if is_channel(url):
		req = session.get(_CHANNELAPI + eplink)
		if verbose and req.status_code == 200:
			print('[DEBUG] Data requested')
			print('[DEBUG] Parsing json API')
		jsdata = req.json()
		title = jsdata['slot']['title']
		hls = jsdata['slot']['chasePlayback']['hls']

		m3u8link = '{x}/{r}/playlist.m3u8'.format(x=hls[:hls.rfind('/')], r=res[:-1])

		if verbose:
			print('[DEBUG] M3U8 Link: {}'.format(m3u8link))
			print('[DEBUG] Title: {}'.format(title))

		return title, m3u8link
	else:
		req = session.get(_PROGRAMAPI + eplink)
		if verbose and req.status_code == 200:
			print('[DEBUG] Data requested')
			print('[DEBUG] Parsing json API')
		jsdata = req.json()
		title = jsdata['series']['title']
		epnum = jsdata['episode']['title']
		hls = jsdata['playback']['hls']
		outputfile = title + ' - ' + epnum
		
		m3u8link = '{x}/{r}/playlist.m3u8'.format(x=hls[:hls.rfind('/')], r=res[:-1])
		
		if verbose:
			print('[DEBUG] M3U8 Link: {}'.format(m3u8link))
			print('[DEBUG] Video title: {}'.format(title))
			print('[DEBUG] Episode number: {}'.format(epnum))
			
		return outputfile, m3u8link


import argparse
import shutil
import sys

from .downloader import *
from .parser import webparse, webparse_m3u8, parsem3u8
from .common import __version__

def main():
    parser = argparse.ArgumentParser(prog='yuu', description='A simple AbemaTV video downloader', epilog='Created by NoAiOne - Version {v}'.format(v=__version__))
    parser.add_argument('--proxies', '-p', required=False, default=None, dest='proxy', help='Use http(s)/socks5 proxies (please add `socks5://` if you use socks5)')
    parser.add_argument('--resolution', '-r', required=False, default='1080p', dest='res', choices=['180p', '240p', '360p', '480p', '720p', '1080p'], help='Resolution (Default: 1080p)')
    parser.add_argument('--output', '-o', required=False, default=None, dest='output', help='Output filename')
    parser.add_argument('--version', '-V', action='version', version='%(prog)s {v} - Created by NoAiOne'.format(v=__version__))
    parser.add_argument('--verbose', '-v', action='store_true', help="Enable verbose")
    parser.add_argument('input', help='AbemaTV url site or m3u8')

    args = parser.parse_args()
    print('[INFO] Starting yuu {ver}...'.format(ver=__version__))

    if args.proxy:
        print('[INFO] Testing proxy')
        sesi = requests.Session()
        sesi.proxies = {'http': args.proxy, 'https': args.proxy}
        # Someebody tell me how to do recursive test properly
        try:
            if args.verbose:
                print('[DEBUG] Testing http+https mode proxy')
            sesi.get('http://httpbin.org/get') # Some test website to check if proxy works or not
            pmode = "HTTP+HTTPS/SOCKS5"
        except:
            if args.verbose:
                print('[DEBUG] Failed')
            sesi = requests.Session()
            sesi.proxies = {'http': args.proxy}
            try:
                if args.verbose:
                    print('[DEBUG] Testing http mode proxy')
                sesi.get('http://httpbin.org/get') # This too but in https mode
                pmode = "HTTP/SOCKS5"
            except:
                if args.verbose:
                    print('[DEBUG] Failed')
                sesi = requests.Session()
                sesi.proxies = {'https': args.proxy} # Final test if it's failed then it will return error
                try:
                    if args.verbose:
                        print('[DEBUG] Testing https mode proxy')
                    sesi.get('http://httpbin.org/get')
                    pmode = "HTTPS/SOCKS5"
                except:
                    if args.verbose:
                        print('[DEBUG] Failed')
                    print('[ERROR] Cannot connect to proxy (Request timeout)')
                    sys.exit(1)
    else:
        sesi = requests.Session()
        try:
            sesi.get('http://httpbin.org/get')
            pmode = "No proxy"
        except:
            print('[ERROR] No connection available to make requests')
            sys.exit(0)
    if args.verbose:
        print('[DEBUG] Using proxy mode: {}'.format(pmode))
		
    print('[INFO] Fetching user token')
    authtoken = getAuthToken(sesi, args.verbose)
    sesi.headers.update({'Authorization': authtoken[0]})
	
    if args.input[-5:] != '.m3u8':
        print('[INFO] Parsing website')
        outputtitle, m3u8link = webparse(args.input, args.res, sesi, args.verbose)
        print('[INFO] Parsing m3u8')
        files, iv, ticket = parsem3u8(m3u8link, sesi, args.verbose)
        if args.output:
            if args.output[-3:] == '.ts':
                output = args.output
            else:
                output = args.output + '.ts'
        else:
            output = '{x} (AbemaTV {r}).ts'.format(x=outputtitle, r=args.res)
        if args.verbose:
            print('[DEBUG] Output file: {}'.format(output))
    elif args.input[-5:] == '.m3u8':
        print('[INFO] Parsing m3u8')
        outputtitle, res = webparse_m3u8(args.input, sesi, args.verbose)
        files, iv, ticket = parsem3u8(args.input, sesi, args.verbose)
        if args.output:
            if args.output[-3:] == '.ts':
                output = args.output
            else:
                output = args.output + '.ts'
        else:
            output = '{x} (AbemaTV {r}).ts'.format(x=outputtitle, r=res)

    print('[INFO] Fetching m3u8 key')
    getkey = fetchVideoKey(ticket, authtoken, sesi, args.verbose)
    
    print('[INFO][DOWN] Starting downloader...')
    dllist, tempdir = getVideo(files, getkey, iv, sesi, args.verbose)
    print('[INFO][DOWN] Finished downloading')
    print('[INFO] Merging video')
    mergeVideo(dllist, output)
    print('[INFO] Finished merging')

    print('[INFO] Cleaning up')
    shutil.rmtree(tempdir)
    sys.exit(0)

if __name__=='__main__':
    main()


import tempfile
from tqdm import tqdm
import requests
import os

from binascii import unhexlify
from Crypto.Cipher import AES
from .parser import fetchVideoKey, getAuthToken, parsem3u8

def decryptData(tsdata, key, iv):
	def _decrypt(d, k, iv):
		dec = AES.new(k, AES.MODE_CBC, IV=iv)
		return dec.decrypt(d)
	
	if iv.startswith('0x'):
		iv = iv[2:]
	iv = unhexlify(iv)
	
	return _decrypt(tsdata, key, iv)
	
def getVideo(fileslist, key, iv, session, verbose):
	print('[INFO][DOWN] Creating temporary folder')
	tempdir = tempfile.mkdtemp()
	dledfiles = []
	
	if not verbose:
		with tqdm(total=len(fileslist), desc='Downloading', ascii=True, unit='file') as pbar:
			for tsf in fileslist:
				outputtemp = os.path.basename(tsf)
				if outputtemp.find('?tver') != -1:
					outputtemp = outputtemp[:outputtemp.find('?tver')]
				outputtemp = tempdir + '\\' + outputtemp
				with open(outputtemp, 'wb') as outf:
					try:
						req = session.get(tsf)
						outf.write(decryptData(req.content, key, iv))
					except Exception as err:
						print('[ERROR] Problem occured\nreason: {}'.format(err))
						import sys; sys.exit(1)
				pbar.update()
				dledfiles.append(outputtemp)
	elif verbose:
		for tsf in fileslist:
			outputtemp = os.path.basename(tsf)
			if outputtemp.find('?tver') != -1:
				outputtemp = outputtemp[:outputtemp.find('?tver')]
			otpt = outputtemp
			outputtemp = os.path.join(tempdir, outputtemp)
			with open(outputtemp, 'wb') as outf:
				try:
					print('[DEBUG][DOWN] Requesting & decrypting content for: {}'.format(otpt))
					req = session.get(tsf)
					outf.write(decryptData(req.content, key, iv))
				except Exception as err:
					print('[ERROR] Problem occured\nreason: {}'.format(err))
					import sys; sys.exit(1)
			dledfiles.append(outputtemp)

	return [dledfiles, tempdir]
	
def mergeVideo(inp, out):
	with open(out, 'wb') as outf:
		with tqdm(total=len(inp), desc='Merging', ascii=True, unit='file') as pbar:
			for i in inp:
				with open(i, 'rb') as c:
					outf.write(c.read())
				os.remove(i)
				pbar.update()


import setuptools
from yuu.common import __version__

with open('README.md', 'r') as f:
    desc = f.read()

setuptools.setup(name = 'yuu',
version = __version__,
description = 'Yuu - A simple AbemaTV video downloader',
long_description = desc,
long_description_content_type = "text/markdown",
author = 'noaione',
author_email = 'noaione0809@gmail.com',
keywords = [
    'ripping', 
    'downloader', 
    'parser'
],
license = 'BSD-3-Clause',
url = 'https://github.com/noaione/yuu',
packages = setuptools.find_packages(),
install_requires = [
    'requests[socks]', 
    'm3u8', 
    'tqdm', 
    'pycryptodome'
],
classifiers = [
    'Development Status :: 5 - Production/Stable', 
    'Programming Language :: Python :: 3', 
    'Programming Language :: Python :: 3.5', 
    'Programming Language :: Python :: 3.6', 
    'Programming Language :: Python :: 3.7'
],
entry_points = {
    'console_scripts': ['yuu=yuu.command:main']
})